import inspect

from snakeguice import inject
from snakeguice.interfaces import Injector
from snakeguice.decorators import GuiceData, GuiceArg, enclosing_frame
from snakeguice.errors import AssistError


def assisted_inject(**kwargs):

    scope = kwargs.get('scope')
    if 'scope' in kwargs:
        del kwargs['scope']

    def _assisted_inject(func):
        if func.__name__ != '__init__':
            raise AssistError('assisted_inject can only be used on __init__s')

        class_locals = enclosing_frame().f_locals

        guice_data = GuiceData.from_class_dict(class_locals)
        guice_data.assisted = True # TODO: I don't like this, but it works for now

        annotations = getattr(func, '__guice_annotations__', {})

        guice_data.init = dict((k, GuiceArg(v, annotations.get(k)))
                                for k, v in kwargs.items())

        return func

    return _assisted_inject


assisted = object()


def AssistProvider(cls):
    guice_data = GuiceData.from_class(cls)
    if not getattr(guice_data, 'assisted', False):
        raise AssistError('AssistProvider can only by used on '
                '@assisted_inject-ed classes')

    class _AssistProvider(object):

        @inject(injector=Injector)
        def __init__(self, injector):
            self._injector = injector

        def get(self):
            return build_factory(self._injector, cls)

    return _AssistProvider


def build_factory(injector, cls):
    guice_data = GuiceData.from_class(cls)

    providers = {}
    for name, guicearg in guice_data.init.items():
        providers[name] = injector.get_provider(guicearg.datatype,
                                                guicearg.annotation)

    all_args = inspect.getargspec(cls.__init__).args[1:]
    needed_args = set(all_args) - set(providers.keys())

    class DynamicFactory(object):

        def create(self, **kwargs):
            if set(kwargs.keys()) - needed_args:
                raise TypeError('TODO: error message here about too many values')

            for name, provider in providers.items():
                kwargs[name] = provider.get()
            return cls(**kwargs)

    return DynamicFactory()



import sys
import inspect


class GuiceData(object):

    def __init__(self):
        self.init = None
        self.methods = {}

    @classmethod
    def from_class(cls, target_class):
        if '__guice__' not in target_class.__dict__:
            guice_data = cls()
            try:
                target_class.__guice__ = guice_data
            except TypeError: # special case for builtin/extension types
                return guice_data
        return target_class.__guice__

    @classmethod
    def composite_from_class(cls, target_class):
        composite_data = GuiceData()
        for _cls in target_class.__mro__[-1::-1]:
            data = GuiceData.from_class(_cls)
            if data.init: # only the last class in the chain
                composite_data.init = data.init
            for name, method in data.methods.items():
                composite_data.methods[name] = method

        return composite_data

    @classmethod
    def from_class_dict(cls, class_dict):
        if '__guice__' not in class_dict:
            class_dict['__guice__'] = cls()
        return class_dict['__guice__']


class GuiceArg(object):

    def __init__(self, datatype=None, annotation=None, scope=None):
        self.datatype = datatype
        self.annotation = annotation
        self.scope = scope

    def __eq__(self, other):
        return (self.datatype, self.annotation, self.scope
                ) == (other.datatype, other.annotation, other.scope)


def _validate_func_args(func, kwargs):
    """Validate decorator args when used to decorate a function."""

    args, varargs, varkw, defaults = inspect.getargspec(func)
    if set(kwargs.keys()) != set(args[1:]): # chop off self
        raise TypeError("decorator kwargs do not match %s()'s kwargs"
                        % func.__name__)


def enclosing_frame(frame=None, level=2):
    """Get an enclosing frame that skips decorator code"""
    frame = frame or sys._getframe(level)
    while frame.f_globals.get('__name__') == __name__: frame = frame.f_back
    return frame


def inject(**kwargs):

    scope = kwargs.get('scope')
    if 'scope' in kwargs:
        del kwargs['scope']

    def _inject(func):
        class_locals = enclosing_frame().f_locals

        guice_data = GuiceData.from_class_dict(class_locals)

        annotations = getattr(func, '__guice_annotations__', {})

        gmethod = dict((k, GuiceArg(v, annotations.get(k)))
                       for k, v in kwargs.items())

        _validate_func_args(func, kwargs)
        if func.__name__ == '__init__':
            guice_data.init = gmethod
        else:
            guice_data.methods[func.__name__] = gmethod

        return func

    return _inject


class annotate(object):

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, method):
        class_locals = enclosing_frame().f_locals
        if '__guice__' in class_locals:
            if method.__name__ in class_locals['__guice__'].methods:
                raise Exception('annotate must be applied before inject')
        method.__guice_annotations__ = self.kwargs
        return method


class provides(object):

    def __init__(self, type):
        self._type = type

    def __call__(self, method):
        method.__guice_provides__ = self._type
        return method



"""Scopes that can be used in the binding process."""

# pylint: disable-msg=C0111

from snakeguice import providers


class _NoScope(object):
    """A default scrope returns the same provider that gets passed in.

    This is internally used and will probably never be directly used in
    a module.
    """

    def scope(self, key, unscoped_provider): # pylint: disable-msg=R0201,W0613
        return unscoped_provider


class _Singleton(object):
    """A singleton scope only allows a single instance to be created for a
    given key.
    """

    def __init__(self):
        self._cached_provider_map = {}

    def scope(self, key, provider):
        cached_provider = self._cached_provider_map.get(key)
        if not cached_provider:
            instance = provider.get()
            cached_provider = self._cached_provider_map[key] = \
                    providers.create_instance_provider(instance)
        return cached_provider


NO_SCOPE = _NoScope()
SINGLETON = _Singleton()



import os
from ConfigParser import SafeConfigParser

from snakeguice.annotation import Annotation


class Config(Annotation):
    """Annotation for ConfigParser style config files."""


class ConfigParserLoader(object):

    def __init__(self, filename):
        self.filename = filename
        self.short_name = os.path.basename(filename)

    def bind_configuration(self, binder):
        parser = SafeConfigParser()
        parser.read(self.filename)
        for section, option, value in _iterate_parser(parser):
            annotation = Config('%s:%s:%s'
                                % (self.short_name, section, option))
            binder.bind(annotation, to_instance=value)


def _iterate_parser(parser):
    for section in parser.sections():
        for option in parser.options(section):
            value = parser.get(section, option)
            yield section, option, value



from snakeguice import errors, providers, scopes


class Key(object):

    def __init__(self, interface, annotation=None):
        self._interface = interface
        self._annotation = annotation

    def __hash__(self):
        return hash((self._interface, self._annotation))

    def __eq__(self, other):
        return (self._interface == other._interface and
                self._annotation == other._annotation)

    def __ne__(self, other):
        return not self == other


class _EmptyBinder(object):

    def get_binding(self, key):
        return None


class Binder(object):

    def __init__(self, parent=None):
        self._parent = parent or _EmptyBinder()
        self._binding_map = {}

    def bind(self, _class, **kwargs):
        key = Key(interface=_class, annotation=kwargs.get('annotated_with'))

        binding = Binding()
        binding.key = key
        binding.scope = kwargs.get('in_scope', scopes.NO_SCOPE)

        if key in self._binding_map:
            raise errors.BindingError('baseclass %r already bound' % _class)

        if 'to' in kwargs:
            if not isinstance(kwargs['to'], type):
                raise errors.BindingError("'to' requires a new-style class")

            binding.provider = providers.create_simple_provider(kwargs['to'])

        elif 'to_provider' in kwargs:
            #TODO: add some validation
            provider = kwargs['to_provider']
            binding.provider = provider

        elif 'to_instance' in kwargs:
            if not isinstance(kwargs['to_instance'], object):
                raise errors.BindingError(
                    "'to_instance' requires an instance of a new-style class")

            provider = kwargs['to_instance']
            binding.provider = providers.create_instance_provider(provider)

        self._binding_map[key] = binding

    def get_binding(self, key):
        return self._binding_map.get(key) or self._parent.get_binding(key)

    def create_child(self):
        return Binder(self)


class LazyBinder(object):

    def __init__(self, parent=None):
        self._parent = parent or _EmptyBinder()
        self._binding_map = {}
        self._errors = []

    def add_error(self, msg):
        import inspect
        frame_rec = inspect.stack()[2]

        location = 'File {0}, line {1}, in {2}'.format(
                frame_rec[1], frame_rec[2], frame_rec[3])
        source = frame_rec[4][frame_rec[5]].strip()
        self._errors.append(BinderErrorRecord(message=msg,
                                              location=location,
                                              source=source))

    @property
    def errors(self):
        return self._errors

    def bind(self, _class, **kwargs):
        key = Key(interface=_class, annotation=kwargs.get('annotated_with'))

        binding = Binding()
        binding.key = key
        binding.scope = kwargs.get('in_scope', scopes.NO_SCOPE)

        if key in self._binding_map:
            self.add_error('baseclass %r already bound' % _class)

        if 'to' in kwargs:
            if not isinstance(kwargs['to'], type):
                self.add_error('to requires a new-style class')

            binding.provider = providers.create_simple_provider(kwargs['to'])

        elif 'to_provider' in kwargs:
            #TODO: add some validation
            provider = kwargs['to_provider']
            binding.provider = provider

        elif 'to_instance' in kwargs:
            if not isinstance(kwargs['to_instance'], object):
                self.add_error('to_instance requires an instance of a '
                               'new-style class')

            provider = kwargs['to_instance']
            binding.provider = providers.create_instance_provider(provider)

        self._binding_map[key] = binding

    def get_binding(self, key):
        return self._binding_map.get(key) or self._parent.get_binding(key)

    def create_child(self):
        return Binder(self)


class Binding(object):

    def __init__(self, key=None, provider=None, scope=None):
        self.key = key
        self.provider = provider
        self.scope = scope


class BinderErrorRecord(object):

    def __init__(self, message, location, source):
        self.message = message
        self.location = location
        self.source = source



"""Errors used by snake-guice"""


class SnakeGuiceError(Exception):
    """Base for all of our errors."""


class BindingError(SnakeGuiceError):
    """Raised when an issue in the binding rules is found."""


class AssistError(SnakeGuiceError):
    """Raised when an issue with assisted injection is found."""


class MultiBindingError(SnakeGuiceError):
    """Raised when a issue with multi-binding is found."""



"""Providers used in the binding process."""

# pylint: disable-msg=C0111
#         providers are so small that we can safely omit doc comments

from snakeguice.decorators import inject
from snakeguice.interfaces import Injector


def create_simple_provider(cls):
    class DynamicSimpleProvider(object):

        @inject(injector=Injector)
        def __init__(self, injector):
            self._injector = injector

        def get(self):
            return self._injector.create_object(cls)

    return DynamicSimpleProvider


def create_instance_provider(obj):
    class DynamicInstanceProvider(object):

        def get(self):
            return obj

    return DynamicInstanceProvider



class Annotation(object):
    """Base class for all annotations."""

    def __init__(self, value):
        self._value = str(value)

    def __hash__(self):
        return hash(self.__class__.__name__ + ":" + self._value)

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def __ne__(self, other):
        return not self == other




class ModuleAdapter(object):
    """Adapts the simple and standard module interfaces to a common one."""

    def __init__(self, module, injector):
        if hasattr(module, 'set_injector'):
            module.set_injector(injector)
        self._module = module

    def configure(self, binder):
        if hasattr(self._module, 'run_configure'):
            self._module.run_configure(binder)
        else:
            self._module.configure(binder)


class Module(object):
    """Base class for all standard modules."""

    def __init__(self):
        self._injector = None

    def install(self, binder, module):
        """Add another module's bindings to a binder."""
        ModuleAdapter(module, self._injector).configure(binder)

    def run_configure(self, binder):
        """A hook for intercepting the configure method. Allows different
        module implementations to change the binder passed to configure.
        """
        self.configure(binder)

    def configure(self, binder):
        """A subclass should override this to configure a binder."""
        raise NotImplementedError

    def set_injector(self, injector):
        self._injector = injector


class _PrivateModuleWrapper(Module):
    """Exists solely to remove the infinite recursion in Private Modules
    caused by the child injector calling run_configure.
    """

    def __init__(self, module):
        self._module = module
        super(_PrivateModuleWrapper, self).__init__()

    def configure(self, binder):
        self._module.configure(binder)


class PrivateModule(Module):
    """Module that uses a child injector to isolate bindings."""

    def expose(self, binder, interface, annotation=None):
        """Expose the child injector to the parent inject for a binding."""
        private_module = self
        class Provider(object):
            def get(self):
                return private_module.private_injector.get_instance(
                        interface, annotation)

        self.original_binder.bind(interface, annotated_with=annotation,
                                  to_provider=Provider)

    def run_configure(self, binder):
        self.original_binder = binder
        private_module = _PrivateModuleWrapper(self)
        self.private_injector = self._injector.create_child(private_module)



import inspect
import warnings

from snakeguice.binder import Binder, Key
from snakeguice.decorators import GuiceData as _GuiceData, inject
from snakeguice.modules import ModuleAdapter
from snakeguice.interfaces import Injector as IInjector


class ProvidesBinderHelper(object):

    def bind_providers(self, module, binder):
        members = [m for m in inspect.getmembers(module)
                   if inspect.ismethod(m[1])]
        for name, method in members:
            if hasattr(method.im_func, '__guice_provides__'):
                type = method.__guice_provides__
                provider = self._build_provider(module, type, method)
                binder.bind(type, to_provider=provider)

    def _build_provider(self, module, type, method):

        class GenericProvider(object):

            @inject(injector=IInjector)
            def __init__(self, injector):
                self._injector = injector

            def get(self):
                kwargs = {}
                method_name = method.__name__
                guice_data = _GuiceData.from_class(module.__class__)
                injectable_args = guice_data.methods.get(method_name, {})
                for name, guicearg in injectable_args.items():
                    kwargs[name] = self._injector.get_instance(
                            guicearg.datatype, guicearg.annotation)
                return method(**kwargs)

        return GenericProvider


class Injector(object):

    def __init__(self, modules=None, binder=None, stage=None):
        if modules is None:
            modules = []
        elif not hasattr(modules, '__iter__'):
            modules = [modules]

        if binder:
            self._binder = binder.create_child()
        else:
            self._binder = Binder()

        self._stage = stage

        self.add_modules(modules)

    def add_modules(self, modules):
        provides_helper = ProvidesBinderHelper()
        for module in modules:
            ModuleAdapter(module, self).configure(self._binder)
            provides_helper.bind_providers(module, self._binder)

    def get_provider(self, cls, annotation=None):
        injector = self

        class DynamicProvider(object):

            def get(self):
                return injector.get_instance(cls, annotation)

        return DynamicProvider()

    def get_instance(self, cls, annotation=None):
        if cls is IInjector: # TODO: i don't like this, but it works for now
            return self

        key = Key(cls, annotation)
        binding = self._binder.get_binding(key)
        if binding:
            provider = binding.scope.scope(key, binding.provider)
            if isinstance(provider, type):
                provider = self.get_instance(provider)
            return provider.get()
        else:
            instance = self.create_object(cls)
            return self.inject_members(instance)

    def create_child(self, modules):
        """Create a new injector that inherits the state from this injector.

        All bindings are inherited. In the future this may become closer to
        child injectors on google-guice.
        """
        binder = self._binder.create_child()
        return Injector(modules, binder=binder, stage=self._stage)

    def create_object(self, cls):
        if not hasattr(cls, '__mro__'):
            warnings.warn("can't create an instance of %r - no __mro__; "
                         "this legacy behavior will be removed in a future "
                          "version")
            return cls

        guice_data = _GuiceData.composite_from_class(cls)

        if not guice_data.init:
            instance = cls()
        else:
            kwargs = {}
            for name, guicearg in guice_data.init.items():
                kwargs[name] = self.get_instance(guicearg.datatype,
                                                 guicearg.annotation)
            instance = cls(**kwargs)

        return instance

    def inject_members(self, instance):
        # this may be a little slow; done twice
        guice_data = _GuiceData.composite_from_class(instance.__class__)

        for name, gm in guice_data.methods.items():
            kwargs = {}
            for param, guicearg in gm.items():
                kwargs[param] = self.get_instance(guicearg.datatype,
                                                  guicearg.annotation)
            getattr(instance, name)(**kwargs)
        return instance


def create_injector(modules):
    """Factory for creating Injector instances."""
    return Injector(modules)



from snakeguice import providers
from snakeguice.annotation import Annotation
from snakeguice.binder import Key
from snakeguice.decorators import inject
from snakeguice.interfaces import Injector
from snakeguice.errors import MultiBindingError


class _MultiBinder(object):

    def __init__(self, binder, interface):
        self._binder = binder
        self._interface = interface
        self._provider = self._get_or_create_provider()

    def _get_or_create_provider(self):
        key = Key(self.multibinding_type(self._interface))
        binding = self._binder.get_binding(key)
        if not binding:
            self._binder.bind(self.multibinding_type(self._interface),
                    to_provider=self._create_provider())
            binding = self._binder.get_binding(key)
        return binding.provider

    def _dsl_to_provider(self, to, to_provider, to_instance):
        if to:
            #TODO: add some validation
            return providers.create_simple_provider(to)
        elif to_provider:
            #TODO: add some validation
            return to_provider
        elif to_instance:
            #TODO: add some validation
            return providers.create_instance_provider(to_instance)
        else:
            raise MultiBindingError('incorrect arguments to %s.add_binding'
                    % self.__class__.__name__)


class List(Annotation):
    """Used for binding lists."""


class ListBinder(_MultiBinder):

    multibinding_type = List

    def add_binding(self, to=None, to_provider=None, to_instance=None):
        provider = self._dsl_to_provider(to, to_provider, to_instance)
        self._provider.add_provider(provider)

    def _create_provider(self):
        class DynamicMultiBindingProvider(object):
            providers = []

            @inject(injector=Injector)
            def __init__(self, injector):
                self._injector = injector

            @classmethod
            def add_provider(cls, provider):
                cls.providers.append(provider)

            def get(self):
                return [self._injector.get_instance(p).get()
                        for p in self.providers]

        return DynamicMultiBindingProvider


class Dict(Annotation):
    """Used for binding dictionaries."""


class DictBinder(_MultiBinder):

    multibinding_type = Dict

    def add_binding(self, key, to=None, to_provider=None, to_instance=None):
        provider = self._dsl_to_provider(to, to_provider, to_instance)
        self._provider.add_provider(key, provider)

    def _create_provider(self):
        binder_self = self

        class DynamicMultiBindingProvider(object):
            providers = {}

            @inject(injector=Injector)
            def __init__(self, injector):
                self._injector = injector

            @classmethod
            def add_provider(cls, key, provider):
                if key in cls.providers:
                    msg = ('duplicate binding for %r in Dict(%s) found'
                            % (key, binder_self.interface.__class__.__name__))
                    raise MultiBindingError(msg)
                cls.providers[key] = provider

            def get(self):
                return dict([(k, self._injector.get_instance(p).get())
                        for k, p in self.providers.items()])

        return DynamicMultiBindingProvider



from os.path import join as joinpath, dirname
from setuptools import setup, Command

from snakeguice import __pkginfo__ as pkg


cmdclass = {}

try:
    from pylint import lint as linter

    class LintCommand(Command):
        description = "Run PyLint over the project"
        user_options = [
                ('report', 'r', 'Print PyLint reports (default: false)'),
                ('rcfile=', 'f', 'PyLint config file'),
        ]

        def initialize_options(self):
            self.report = False
            self.rcfile = None

        def finalize_options(self):
            if not self.rcfile:
                self.rcfile = joinpath(dirname(__file__), 'pylint.cfg')
            self.report = bool(self.report)

        def run(self):
            args = ['--rcfile', self.rcfile]
            if not self.report:
                args.append('-rn')

            ##pkgs = list(set(p.split(".")[0] for p in options.setup.packages))
            pkgs = ['snakeguice']
            linter.Run(args + pkgs)

    cmdclass['lint'] = LintCommand

except ImportError:
    """Pylint is not installed."""

try:
    import nose

    class TestsCommand(Command):
        description = "Run unit tests with nose"
        user_options = [('coverage', 'c', 'Generate figleaf sections.')]

        def initialize_options(self):
            self.coverage = None

        def finalize_options(self):
            self.coverage = bool(self.coverage)

        def run(self):
            if self.coverage:
                import figleaf
                figleaf.start()
            nose.run(argv=['nosetests'])
            if self.coverage:
                figleaf.stop()
                figleaf.write_coverage('.figleaf')

    cmdclass['tests'] = TestsCommand

except ImportError:
    """Nose is not installed."""

try:
    import mote.runner

    class SpecsCommand(Command):
        description = "Run specs with mote"
        user_options = [('quiet', 'q', "Don't print the specification")]

        def initialize_options(self):
            self.quiet = None

        def finalize_options(self):
            self.quiet = bool(self.quiet)

        def run(self):
            sys.argv.append('specs')
            mote.runner.main()

    cmdclass['specs'] = SpecsCommand

except ImportError:
    """mote is not installed."""

setup(
    name=pkg.libname,
    version=pkg.version,
    license=pkg.license,
    packages=pkg.packages,
    author=pkg.author,
    author_email=pkg.author_email,
    description=pkg.short_desc,
    long_description=pkg.long_desc,
    classifiers=pkg.classifiers,
    url=pkg.url,
    cmdclass=cmdclass
)




class ParameterInterceptor(object):

    def __init__(self, injector):
        self._injector = injector

    def __call__(self, annotation=None, **kwargs):
        #TODO: add a test for more than one kwargs
        (param_name, param_type) = kwargs.items()[0]

        def callback(method):
            def _callback(*args, **kwargs):
                kwargs[param_name] = self._injector.get_instance(
                        param_type, annotation)
                return method(*args, **kwargs)
            return _callback

        return callback



"""Main API entry point for snake-guice."""

from snakeguice.injector import create_injector, Injector
from snakeguice.decorators import inject, annotate, provides
from snakeguice.errors import SnakeGuiceError, BindingError
from snakeguice.interceptors import ParameterInterceptor



"""snake-guice packaging information"""
# pylint: disable-msg=C0103,W0622

libname = 'snake-guice'
modname = 'snakeguice'
packages = ['snakeguice']

version = '0.4.1'

license = 'MIT'
copyright = 'Copyright (c) 2008-2010 David Stanek (dstanek@dstanek.com)'

short_desc = 'A simple, lightweight Python dependency injection framework'
long_desc = """\
 snake-guice is a simple, lightweight Python dependency injection
 framework based on google-guice. The Guice way to do things is quite a
 bit different than the current breed of XML IoC containers."""

author = 'David Stanek'
author_email = 'dstanek@dstanek.com'

url = 'https://github.com/dstanek/snake-guice'
mailing_list = 'http://groups.google.com/group/snake-guice'

classifiers =  [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
]
